/*
 * src/netapplet.c - the network control applet
 *
 * Copyright (C) 2004 Novell, Inc.
 *
 * Licensed under the GNU GPL v2.  See COPYING
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <errno.h>
#include <gnome.h>
#include <math.h>
#include <glade/glade-xml.h>
#include <gnome-keyring.h>

#include "eggtrayicon.h"
#include "netcommon.h"

#ifdef ENABLE_NLS
# include <libintl.h>
# define _(String) gettext (String)
# ifdef gettext_noop
#   define N_(String) gettext_noop (String)
# else
#   define N_(String) (String)
# endif
#else
# define _(String)  (String)
# define N_(String) (String)
#endif

#define ETHERNET_ICON		"gnome-dev-ethernet"
#define DIALUP_ICON		"gnome-dev-modem"
#define WIRELESS_ICON_0         "wireless-tower-0"
#define WIRELESS_ICON_1         "wireless-tower-1"
#define WIRELESS_ICON_2         "wireless-tower-2"
#define WIRELESS_ICON_3         "wireless-tower-3"
#define WIRELESS_ICON_4         "wireless-tower-4"
#define DISCONNECT_ICON		"stock_calc-cancel"
#define BROKEN_ICON		"stock_calc-cancel"
#define UNKNOWN_ICON            "stock_unknown"

#define GLADE_APPLET		GLADEDIR "/netapplet.glade"

#define YAST2_ARGV { "/opt/kde3/bin/kdesu", "-n", "/sbin/yast2", "lan", NULL }

typedef struct {
	char *interface;	/* interface name, e.g. eth0 */
	const char *type;	/* TYPE_{DIALUP,ETHERNET,WIRELESS} */
	const char *icon;	/* icon id */
} Connection;

typedef struct {
	double strength;	/* wireless quality, should be [0.0, 1.0) */
	char *essid;		/* ESSID name */
	gboolean is_encrypted;	/* is this network encrypted ? */
} Accesspoint;

typedef struct {
	EggTrayIcon *tray_icon;
	GtkWidget *button;
	GtkWidget *icon;
	GtkTooltips *tooltips;
	GtkWidget *popup_menu;
	const char *icon_name;
	int last_panel_h;

	GSList *connections;		/* of Connection */
	Connection *active;		/* active connection */

	GSList *accesspoints;		/* of Accesspoints */
	Accesspoint *active_ap;         /* Active access point */
	double link_quality;		/* global link quality [0,1] */
	guint update_tag;
	guint update_freeze;
} NetApplet;

static NetApplet *netapplet;

static void populate_popup_menu (void);
static void netapplet_handle_active (char **args);
static void netapplet_handle_accesspoints (GSList *aps);
static void netapplet_handle_wireless (char **args);

static gboolean disconnected = FALSE;
static void
netapplet_set_icon (const char *icon_name)
{
	GdkPixbuf *pixbuf;
	char *tip;
	int panel_w;
	int panel_h;
	int icon_size;

	if (strcmp(icon_name, DISCONNECT_ICON))
		disconnected = FALSE;

	gtk_window_get_size (
		GTK_WINDOW (gtk_widget_get_toplevel (netapplet->button)),
		&panel_w, &panel_h);

	if (netapplet->icon_name != NULL &&
	    strcmp (netapplet->icon_name, icon_name) == 0 &&
	    panel_h == netapplet->last_panel_h)
		return;

	netapplet->icon_name = icon_name;
	netapplet->last_panel_h = panel_h;

	if (panel_h < 30)
		icon_size = 16;
	else
		icon_size = 24;

	pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
					   icon_name, icon_size, 0, NULL);
	
	/*
	 * Scale the icon rather than clip it if our allocation just isn't
	 * what we want it to be.
	 */
	if (GTK_WIDGET_REALIZED (netapplet->icon) &&
	    netapplet->icon->allocation.height < icon_size) {
		int new_size = netapplet->icon->allocation.height;
		GdkPixbuf *new_pixbuf;

		new_pixbuf = gdk_pixbuf_scale_simple (pixbuf,
						      new_size, new_size,
						      GDK_INTERP_BILINEAR);

		g_object_unref (pixbuf);
		pixbuf = new_pixbuf;
	}

	gtk_image_set_from_pixbuf (GTK_IMAGE (netapplet->icon), pixbuf);
	g_object_unref (pixbuf);

	if (netapplet->tooltips == NULL)
		netapplet->tooltips = gtk_tooltips_new ();

	if (netapplet->active == NULL)
		tip = g_strdup (_("Disconnected"));
	else if (strcmp (netapplet->active->type, "Ethernet") == 0)
		tip = g_strdup (_("Ethernet connection"));
	else if (strcmp (netapplet->active->type, "Dialup") == 0)
		tip = g_strdup (_("Dial-up connection"));
	else if (strcmp (netapplet->active->type, "Wireless") == 0) {
		if (netapplet->active_ap == NULL)
			tip = g_strdup ("Wireless connection");
		else if (!netapplet->active_ap->is_encrypted) {
			tip = g_strdup_printf ("Wireless connection: %d%%",
				(int)(netapplet->active_ap->strength * 100));
		} else
			tip = g_strdup_printf ("Wireless connection (secure): "
			 "%d%%", (int)(netapplet->active_ap->strength * 100));
	} else {
		g_warning ("icon_name=%s\n", icon_name);
		g_assert_not_reached ();
		return; /* bleh */
	}

	gtk_tooltips_set_tip (netapplet->tooltips, netapplet->button,
			      tip, NULL);
	g_free (tip);
}

static void
netapplet_set_connection (Connection *conn)
{
	char *args[2];
	args[0] = NULL;
	args[1] = conn->interface;

	netapplet_handle_active(args);
}

static void
conn_free (Connection *conn)
{
	g_free (conn->interface);
	g_free (conn);
}

static void
netapplet_free_connections (void)
{
	g_slist_foreach (netapplet->connections, (GFunc) conn_free, NULL);
	g_slist_free (netapplet->connections);
	netapplet->connections = NULL;
	netapplet->active = NULL;
}

static char *
wireless_icon (Connection *conn)
{
	float strength;
	gboolean encrypted;

	if (strcmp (conn->type, "Wireless") != 0 ||
	    netapplet->active_ap == NULL)
		return WIRELESS_ICON_0;

	encrypted = netapplet->active_ap->is_encrypted;
	strength = netapplet->link_quality;

	if (strength < 0.25)
		return encrypted ? WIRELESS_ICON_1 : WIRELESS_ICON_1;

	if (strength < 0.50)
		return encrypted ? WIRELESS_ICON_2 : WIRELESS_ICON_2;

	if (strength < 0.75)
		return encrypted ? WIRELESS_ICON_3 : WIRELESS_ICON_3;

	return encrypted ? WIRELESS_ICON_4 : WIRELESS_ICON_4;
}

static void
netapplet_get_active (void)
{
	FILE *fp = fopen("/etc/gnet/lastiface", "r");
	char buffer[512];
	char *args[2];
	memset(buffer, 0, 512);
	
	if (fp == NULL) {
		system("gnet init");
		netapplet_get_active();
		return;
	}

	while (fgets(buffer, 511, fp) != NULL) {
		if (strstr(buffer, "\n"))
			*strstr(buffer, "\n") = '\0';
		args[0] = NULL;
		args[1] = buffer;
		netapplet_handle_active(args);
		memset(buffer, 0, 512);
	}
	fclose(fp);
}

static int
netapplet_get_connections (void *data)
{
	FILE *fp = NULL;
	char buffer[512];
	Connection *gprsconn = NULL;

	if (netapplet->update_tag == 0) {
		netapplet->update_tag = gtk_timeout_add(1000, 
				netapplet_get_connections, NULL);
		netapplet->update_freeze = FALSE;	
	}
	if (netapplet->update_freeze)
		return TRUE;
	
	fp = popen("/etc/gnet/gnet_get_ifaces", "r");
	gprsconn = g_new0(Connection, 1);

	memset(buffer, 0, 512);
	
	netapplet_free_connections ();
	while (fgets(buffer, 511, fp) != NULL) {
		Connection *conn = g_new0(Connection, 1);
		char **args = NULL;
		
		if (strstr(buffer, "\n"))
			*strstr(buffer, "\n") = '\0';
		
		args = g_strsplit(buffer, " ", -1);
		
		conn->interface = g_strdup (args[0]);

		if (!strcmp(args[1], "wireless")) {
			conn->type = "Wireless";
			conn->icon = wireless_icon (conn);
		}
		else if (!strcmp(args[1], "wired")) {
			conn->type = "Ethernet";
			conn->icon = ETHERNET_ICON;
		}
		else {
			conn->type = "Unknown";
			conn->icon = UNKNOWN_ICON;
		}
		netapplet->connections = g_slist_append(netapplet->connections, conn);
		memset(buffer, 0, 512);
		g_strfreev(args);
	}
	gprsconn->interface=g_strdup("ppp0");
	gprsconn->type="Dialup";
	gprsconn->icon=DIALUP_ICON;
	netapplet->connections = g_slist_append(netapplet->connections, gprsconn);
	pclose(fp);
	populate_popup_menu();
	netapplet_get_active ();
	return TRUE;
}

static void
netapplet_get_wireless ()
{
	FILE *fp = fopen("/etc/gnet/lastap", "r");
	char buffer[512];
	char *args[2];
	memset(buffer, 0, 512);

	if (fp == NULL) {
		system("gnet init");
		netapplet_get_wireless();
		return;
	}
	
	while (fgets(buffer, 511, fp) != NULL) {
		GSList *aps = netapplet->accesspoints;
		if (strstr(buffer, "\n"))
			*strstr(buffer, "\n") = '\0';
		args[0] = NULL;
		args[1] = buffer;
		args[3] = "0.000000";

		while (aps) {
			Accesspoint *ap = (Accesspoint *)aps->data;
			if (!strcmp(ap->essid, buffer)) {
				args[2] = ap->is_encrypted?"1":"0";
			}
			aps = aps->next;
		}
		
		netapplet_handle_wireless(args);
		memset(buffer, 0, 512);
	}
	fclose(fp);
	
}

static void
netapplet_get_accesspoints ()
{
	FILE *fp = popen("list_aps_short", "r");
	char buffer[512];
	GSList *aps = NULL;
	
	memset(buffer, 0, 512);
	while (fgets(buffer, 511, fp) != NULL) {
		char **args;
		char **strength;
		int str, max;
		Accesspoint *ap;
		if (strstr(buffer, "\n"))
			*(strstr(buffer, "\n")) = '\0';
		
		args = g_strsplit(buffer, " ", -1);
		
		ap = g_new (Accesspoint, 1);
		ap->essid = g_strdup (args[0]);
		ap->is_encrypted = !(strcmp(args[3], "on"));
		
		strength = g_strsplit(args[4], "/", -1);
		str = atoi(strength[0]);
		max = atoi(strength[1]);
		ap->strength = log(str)/log(max);
		g_strfreev(strength);
		
		aps = g_slist_append(aps, ap);
		
		g_strfreev(args);
		memset(buffer, 0, 512);
	}
	pclose(fp);

	netapplet_handle_accesspoints(aps);
	g_slist_free (aps);
}

static void
ap_free (Accesspoint *ap)
{
	g_free (ap->essid);
	g_free (ap);
}

static void
netapplet_free_accesspoints (void)
{
	g_slist_foreach (netapplet->accesspoints, (GFunc) ap_free, NULL);
	g_slist_free (netapplet->accesspoints);
	netapplet->accesspoints = NULL;
	netapplet->active_ap = NULL;
}

static Connection *
netapplet_get_connection_by_interface (const char *interface)
{
	GSList *iter;

	for (iter = netapplet->connections; iter != NULL; iter = iter->next) {
		Connection *conn = iter->data;

		if (strcmp (conn->interface, interface) == 0)
			return conn;
	}

	return NULL;
}

static Accesspoint *
netapplet_get_accesspoint_by_essid (const char *essid)
{
	GSList *iter;

	for (iter = netapplet->accesspoints; iter != NULL; iter = iter->next) {
		Accesspoint *ap = iter->data;

		if (strcmp (ap->essid, essid) == 0)
			return ap;
	}

	return NULL;
}

static void
netapplet_connect (void)
{
	g_return_if_fail (netapplet != NULL);
	g_return_if_fail (netapplet->tray_icon != NULL);

	netapplet_get_connections (NULL);
}

static void
connection_activate_cb (GtkMenuItem *mi G_GNUC_UNUSED,
			gpointer user_data G_GNUC_UNUSED)
{
	const char *interface;
	Connection *conn;
	char *cmdline;
	
	interface = g_object_get_data (G_OBJECT (mi), "interface");
	conn = netapplet_get_connection_by_interface (interface);
	g_assert (conn != NULL);

	cmdline = g_strdup_printf("gnet setiface=%s setconfig=nochange setap=nochange",
			interface);
	system(cmdline);
	g_free(cmdline);
	netapplet_set_connection (conn);
}

static void
essid_activate_cb (GtkMenuItem *mi G_GNUC_UNUSED,
		   gpointer user_data G_GNUC_UNUSED)

{
	const char *essid;
	Accesspoint *ap;
	char *cmdline = NULL;
	essid = g_object_get_data (G_OBJECT (mi), "essid");
	ap = netapplet_get_accesspoint_by_essid (essid);
	g_assert (ap);

	cmdline = g_strdup_printf(
		"gnet setiface=nochange setconfig=nochange setap=%s", essid);
	system(cmdline);
	g_free(cmdline);
}

static void
other_essid_cb (GtkMenuItem *mi G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
{
	char *cmdline = NULL;

	cmdline = g_strdup_printf(
		"gnet setiface=nochange setconfig=nochange setap=Other");

	system(cmdline);
	g_free(cmdline);
}

static GtkWidget *
get_label (GtkWidget *info_dialog, GladeXML *xml, const char *name)
{
	GtkWidget *label;

	if (xml != NULL) {
		label = glade_xml_get_widget (xml, name);
		g_object_set_data (G_OBJECT (info_dialog), name, label);
	} else
		label = g_object_get_data (G_OBJECT (info_dialog), name);

	return label;
}

static void
show_info_cb (GtkMenuItem *mi G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
{
	static GtkWidget *info_dialog = NULL;
	char *addr = NULL, *mask = NULL, *broadcast = NULL;
	char *dest = NULL, *mac = NULL;
	GtkWidget *label;
	GladeXML *xml = NULL;
	Connection *conn;
	struct ifreq ifr;
	int fd, flags;

	conn = netapplet->active;
	if (!conn) {
		GtkWidget *error_dialog;

		error_dialog = gtk_message_dialog_new_with_markup (
				NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				_("<span weight=\"bold\" size=\"larger\">"
				  "Error displaying connection information: "
				  "</span>\n\n"
				  "No active connection!"));
		gtk_dialog_run (GTK_DIALOG (error_dialog));
		gtk_widget_destroy (error_dialog);
		return;
	}

	fd = socket (AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		GtkWidget *error_dialog;

		error_dialog = gtk_message_dialog_new_with_markup (NULL, 0,
				GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				_("<span weight=\"bold\" size=\"larger\">"
				  "Error displaying connection information: "
				"</span>\n\nUnable to open socket!"));
		gtk_dialog_run (GTK_DIALOG (error_dialog));
		gtk_widget_destroy (error_dialog);
		return;
	}

	ifr.ifr_addr.sa_family = AF_INET;

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (ioctl (fd, SIOCGIFADDR, &ifr) == 0)
		addr = g_strdup (inet_ntoa (((struct sockaddr_in *)
					&ifr.ifr_addr)->sin_addr));

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (ioctl (fd, SIOCGIFFLAGS, &ifr) < 0) {
		GtkWidget *error_dialog;

		error_dialog = gtk_message_dialog_new_with_markup (NULL, 0,
				GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
				_("<span weight=\"bold\" size=\"larger\">"
				  "Error displaying information: "
				  "</span>\n\n"
				  "SIOCGIFFLAGS failed on socket!"));
		gtk_dialog_run (GTK_DIALOG (error_dialog));
		gtk_widget_destroy (error_dialog);
		goto out;
	}
	flags = ifr.ifr_flags;

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (flags & IFF_BROADCAST && ioctl (fd, SIOCGIFBRDADDR, &ifr) == 0)
		broadcast = g_strdup (inet_ntoa (((struct sockaddr_in *)
					&ifr.ifr_broadaddr)->sin_addr));

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (ioctl (fd, SIOCGIFNETMASK, &ifr) == 0)
		mask = g_strdup (inet_ntoa (((struct sockaddr_in *)
					&ifr.ifr_addr)->sin_addr));

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (flags & IFF_POINTOPOINT && ioctl (fd, SIOCGIFDSTADDR, &ifr) == 0)
		dest = g_strdup (inet_ntoa (((struct sockaddr_in *)
					&ifr.ifr_dstaddr)->sin_addr));

	g_strlcpy (ifr.ifr_name, conn->interface, sizeof (ifr.ifr_name));
	if (ioctl (fd, SIOCGIFHWADDR, &ifr) == 0)
		mac = g_strdup_printf ("%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
		       (unsigned char) ifr.ifr_hwaddr.sa_data[0],
		       (unsigned char) ifr.ifr_hwaddr.sa_data[1],
		       (unsigned char) ifr.ifr_hwaddr.sa_data[2],
		       (unsigned char) ifr.ifr_hwaddr.sa_data[3],
		       (unsigned char) ifr.ifr_hwaddr.sa_data[4],
		       (unsigned char) ifr.ifr_hwaddr.sa_data[5]);

	if (!info_dialog) {
		xml = glade_xml_new (GLADE_APPLET, "dialog-information", NULL);
		info_dialog = glade_xml_get_widget (xml, "dialog-information");
	}

	label = get_label (info_dialog, xml, "label-interface");
	gtk_label_set_text (GTK_LABEL (label), conn->interface);

	label = get_label (info_dialog, xml, "label-type");
	gtk_label_set_text (GTK_LABEL (label), conn->type);

	label = get_label (info_dialog, xml, "label-ip-address");
	gtk_label_set_text (GTK_LABEL (label), addr);

	label = get_label (info_dialog, xml, "label-destination-address");
	if (flags & IFF_POINTOPOINT) {
		gtk_label_set_text (GTK_LABEL (label), dest);
		gtk_widget_show (label);
	} else
		gtk_widget_hide (label);

	label = get_label (info_dialog, xml,
			   "label-destination-address-label");
	if (flags & IFF_POINTOPOINT) {
		gtk_label_set_text (GTK_LABEL (label), dest);
		gtk_widget_show (label);
	} else
		gtk_widget_hide (label);

	label = get_label (info_dialog, xml, "label-broadcast-address");
	gtk_label_set_text (GTK_LABEL (label), broadcast);

	label = get_label (info_dialog, xml, "label-subnet-mask");
	gtk_label_set_text (GTK_LABEL (label), mask);

	label = get_label (info_dialog, xml, "label-hardware-address");
	gtk_label_set_text (GTK_LABEL (label), mac);

	if (xml != NULL) {
		gtk_dialog_run (GTK_DIALOG (info_dialog));
		gtk_widget_destroy (info_dialog);
		info_dialog = NULL;
		g_object_unref (xml);
	} else
		gtk_window_present (GTK_WINDOW (info_dialog));

out:
	close (fd);
	g_free (addr);
	g_free (broadcast);
	g_free (mask);
	g_free (dest);
	g_free (mac);
}

static void
configure_activate_cb (GtkMenuItem *mi G_GNUC_UNUSED,
		       gpointer user_data G_GNUC_UNUSED)
{
	system("gnet setiface=nochange setap=nochange");
}

static void
quit_cb (GtkMenuItem *mi G_GNUC_UNUSED,
	 gpointer user_data G_GNUC_UNUSED)
{
	gtk_timeout_remove(netapplet->update_tag);
	gtk_exit (0);
}

static void
remove_old_items (gpointer node, gpointer user_data)
{
	gtk_container_remove (GTK_CONTAINER (user_data), GTK_WIDGET (node));
}

static GtkWidget *
percentage_menu_item_new (const char *text, double percentage, gboolean locked)
{
	GtkWidget *menu_item;
	GtkWidget *box, *label, *progress;
	PangoContext *context;
	PangoFontMetrics *metrics;
	int ascent;

	menu_item = gtk_image_menu_item_new ();

	if (locked) {
		GtkWidget *image;

		image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_AUTHENTICATION,
						  GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item),
					       image);
	}

	box = gtk_hbox_new (FALSE, 6);
	gtk_container_add (GTK_CONTAINER (menu_item), box);
	gtk_widget_show (box);

	label = gtk_label_new (text);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 0);
	gtk_widget_show (label);

	progress = gtk_progress_bar_new ();
	gtk_progress_set_percentage (GTK_PROGRESS (progress), percentage);

	g_object_set_data (G_OBJECT (menu_item), "progress_widget", progress);

	context = gtk_widget_get_pango_context (progress);
	metrics = pango_context_get_metrics (context,
			pango_context_get_font_description (context), NULL);

	ascent = pango_font_metrics_get_ascent (metrics) * 1.5 / PANGO_SCALE;
	gtk_widget_set_size_request (progress, ascent * 5, ascent);

	gtk_box_pack_end (GTK_BOX (box), progress, FALSE, TRUE, 0);
	gtk_widget_show (progress);

	return menu_item;
}

static void
populate_popup_menu (void)
{
	GtkWidget *mi, *image;
	GSList *iter;

	g_list_foreach (GTK_MENU_SHELL (netapplet->popup_menu)->children,
			remove_old_items, netapplet->popup_menu);

	/* Build list of network connections ... */
	if (netapplet->connections)
		mi = gtk_menu_item_new_with_label (_("Network Connections"));
	else
		mi = gtk_menu_item_new_with_label (
					_("No network connections available"));

	gtk_widget_set_sensitive (mi, FALSE);
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);

	for (iter = netapplet->connections; iter != NULL; iter = iter->next) {
		Connection *conn = iter->data;
		GdkPixbuf *pixbuf;
		char *display_name;

		if (conn == netapplet->active)
			display_name = g_strdup_printf (_("%s: %s (active)"),
							conn->type,
							conn->interface);
		else
			display_name = g_strdup_printf ("%s: %s", conn->type,
							conn->interface);
		mi = gtk_image_menu_item_new_with_label (display_name);
		g_free (display_name);

		pixbuf = gtk_icon_theme_load_icon (
				gtk_icon_theme_get_default (),
				conn->icon, 16, 0, NULL);
		image = gtk_image_new_from_pixbuf (pixbuf);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mi),
					       image);
		gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu),
				       mi);
		g_object_set_data_full (G_OBJECT (mi), "interface",
					g_strdup (conn->interface), g_free);
		g_signal_connect (mi, "activate",
				  G_CALLBACK (connection_activate_cb), NULL);

		gtk_widget_set_sensitive (mi, TRUE);
		gtk_widget_show (mi);
	}

	mi = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);

	/* Build list of wireless accesspoints ... */
	if (netapplet->accesspoints)
		mi = gtk_menu_item_new_with_label (_("Wireless Networks"));
	else if (!(netapplet->active &&
		   strcmp (netapplet->active->type, "Wireless") == 0))
		mi = gtk_menu_item_new_with_label (_("Wireless disabled"));
	else
		mi = gtk_menu_item_new_with_label (
					_("No wireless networks available"));
	gtk_widget_set_sensitive (mi, FALSE);
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);

	for (iter = netapplet->accesspoints; iter != NULL; iter = iter->next) {
		Accesspoint *ap = iter->data;
		char *display_name;
		float strength = ap->strength;

		if (ap == netapplet->active_ap) {
			display_name = g_strdup_printf (_("%s (active)"),
							ap->essid);
			netapplet->link_quality = strength;
		} else
			display_name = g_strdup_printf ("%s", ap->essid);

		mi = percentage_menu_item_new (display_name, strength,
					       ap->is_encrypted);
		gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu),
				       mi);
		g_object_set_data_full (G_OBJECT (mi), "essid",
					g_strdup (ap->essid), g_free);
		g_signal_connect (mi, "activate",
				  G_CALLBACK (essid_activate_cb), NULL);
		gtk_widget_set_sensitive (mi, TRUE);
		gtk_widget_show (mi);
		g_free (display_name);
	}

	if (netapplet->active &&
	    strcmp (netapplet->active->type, "Wireless") == 0) { 
		mi = gtk_menu_item_new_with_label ("Other ...");
		gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu),
				       mi);
		g_signal_connect (mi, "activate", G_CALLBACK (other_essid_cb),
				  NULL);

		gtk_widget_set_sensitive (mi, TRUE);
		gtk_widget_show (mi);
	}

	mi = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);

	/* 'Connection Information' item */
	image = gtk_image_new_from_stock (GTK_STOCK_PROPERTIES,
					  GTK_ICON_SIZE_MENU);
	mi = gtk_image_menu_item_new_with_mnemonic (
					_("Connection _Information"));
	gtk_image_menu_item_set_image ((GtkImageMenuItem *) mi, image);
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);
	g_signal_connect (mi, "activate", G_CALLBACK (show_info_cb), netapplet);

	/* 'Configure Network Settings' item */
	image = gtk_image_new_from_stock (GTK_STOCK_PREFERENCES,
					  GTK_ICON_SIZE_MENU);
	mi = gtk_image_menu_item_new_with_mnemonic (
				_("_Configure Network Settings"));
	gtk_image_menu_item_set_image ((GtkImageMenuItem *) mi, image);
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);
	g_signal_connect (mi, "activate", G_CALLBACK (configure_activate_cb),
			  netapplet);

	mi = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);

	mi = gtk_image_menu_item_new_with_mnemonic (_("_Remove From Panel"));
	image = gtk_image_new_from_stock (GTK_STOCK_REMOVE,
					  GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mi), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (netapplet->popup_menu), mi);
	gtk_widget_show (mi);
	g_signal_connect (mi, "activate", G_CALLBACK (quit_cb), NULL);
}

static Accesspoint *
get_accesspoint_for_essid (char *essid)
{
	GSList *iter;

	for (iter = netapplet->accesspoints; iter != NULL; iter = iter->next) {
		Accesspoint *ap = iter->data;

		if (strcmp (ap->essid, essid) == 0)
			return ap;
	}

	return NULL;
}

static void
update_popup_accesspoint_strengths (void)
{
	GList *iter;

	for (iter = GTK_MENU_SHELL (netapplet->popup_menu)->children;
	     iter != NULL; iter = iter->next) {
		GtkWidget *mi = iter->data;
		char      *essid;

		essid = g_object_get_data (G_OBJECT (mi), "essid");
		if (essid) {
			Accesspoint *ap;
			GtkWidget *progress;

			ap = get_accesspoint_for_essid (essid);
			if (ap == NULL)
				continue;

			progress = g_object_get_data (G_OBJECT (mi),
						      "progress_widget");
			if (progress == NULL)
				continue;

			gtk_progress_set_percentage (GTK_PROGRESS (progress),
						     ap->strength);
		}
	}
}

static void
menu_position_func (GtkMenu *menu G_GNUC_UNUSED, int *x, int *y,
		    gboolean *push_in, gpointer user_data G_GNUC_UNUSED)
{
	GdkScreen *screen;
	int screen_w, screen_h;
	int button_x, button_y;
	int panel_w, panel_h;
	GtkRequisition requisition;

	screen = gtk_widget_get_screen (netapplet->button);
	screen_w = gdk_screen_get_width (screen);
	screen_h = gdk_screen_get_height (screen);

	gdk_window_get_origin (netapplet->button->window,
			       &button_x, &button_y);

	gtk_window_get_size (
		GTK_WINDOW (gtk_widget_get_toplevel (netapplet->button)),
		&panel_w, &panel_h);

	*x = button_x;

	/*
	 * Check to see if we would be placing the menu off the
	 * end of the screen.
	 */
	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);
	
	if (button_y + panel_h + requisition.height >= screen_h)
		*y = button_y - requisition.height;
	else
		*y = button_y + panel_h;

	*push_in = TRUE;
}

static void
button_toggled_cb (GtkToggleButton *button,
		   gpointer user_data G_GNUC_UNUSED)
{
	netapplet->update_freeze = gtk_toggle_button_get_active(button);
	gtk_menu_popup (GTK_MENU (netapplet->popup_menu), NULL, NULL,
			menu_position_func, NULL, 0,
			gtk_get_current_event_time ());
}

static void
menu_deactivate_cb (GtkMenuShell *menu_shell G_GNUC_UNUSED,
		    gpointer user_data G_GNUC_UNUSED)
{
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (netapplet->button),
				      FALSE);
	gtk_menu_popdown (GTK_MENU (netapplet->popup_menu));
}

static gboolean
accesspoints_differ (Accesspoint *ap1, Accesspoint *ap2)
{
	if (ap1 == NULL || ap2 == NULL)
		return ap1 != ap2;
		
	if (strcmp (ap1->essid, ap2->essid) != 0)
		return TRUE;

	if (ap1->is_encrypted != ap2->is_encrypted)
		return TRUE;

	return FALSE;
}

static gboolean
accesspoint_in_list (Accesspoint *ap, GSList *list)
{
	GSList *iter;

	for (iter = list; iter != NULL; iter = iter->next) {
		Accesspoint *ap2 = iter->data;

		if (! accesspoints_differ (ap, ap2))
			return TRUE;
	}

	return FALSE;
}

static gboolean
accesspoint_lists_differ (GSList *aps1, GSList *aps2)
{
	GSList *iter;

	if (g_slist_length (aps1) != g_slist_length (aps2))
		return TRUE;

	for (iter = aps1; iter != NULL; iter = iter->next) {
		Accesspoint *ap = iter->data;

		if (!accesspoint_in_list (ap, aps2))
			return TRUE;
	}

	return FALSE;
}

static void
netapplet_handle_accesspoints (GSList *aps)
{
	GSList *new_aps = NULL;
	Accesspoint *new_active = NULL;
	gboolean changed = FALSE;
	GSList *cur = aps;
	
	for (cur = aps; cur; cur = cur->next) {
		Accesspoint *ap = (Accesspoint *)cur->data;

		if (netapplet->active_ap != NULL &&
		    strcmp (netapplet->active_ap->essid, ap->essid) == 0)
			new_active = ap;

		new_aps = g_slist_append (new_aps, ap);
	}

	if (accesspoint_lists_differ (netapplet->accesspoints, new_aps) ||
	    accesspoints_differ (netapplet->active_ap, new_active))
		changed = TRUE;

	netapplet_free_accesspoints ();
	netapplet->accesspoints = new_aps;
	netapplet->active_ap = new_active;

	netapplet->active->icon = wireless_icon (netapplet->active);
	netapplet_set_icon (netapplet->active->icon);

	if (changed)
		populate_popup_menu ();
	else
		update_popup_accesspoint_strengths ();
}

/*
 * handle the 'wireless' message, which tells us the active ESSID
 *
 * args[0] is 'wireless' (ignored)
 * args[1] is the ESSID
 * args[2] is 1 if encrypted is enabled and 0 otherwise
 * args[3] is the link quality [0, 1]
 */
static void
netapplet_handle_wireless (char **args)
{
	char *new_icon;
	GSList *iter;

	//netapplet->link_quality = strtod (args[3], NULL);
	new_icon = wireless_icon (netapplet->active);
	if (netapplet->active->icon != new_icon) {
		netapplet->active->icon = new_icon;
		netapplet_set_icon (netapplet->active->icon);
		goto out;	/* we want to run populate_popup_menu () */
	}

	/*
	 * Bail out early if we see that our active essid has not changed.
	 * We handle the link quality update above, so there is nothing
	 * else to do.  This avoids redrawing the menu, too.
	 */
	if (netapplet->active_ap &&
	    strcmp (args[1], netapplet->active_ap->essid) == 0)
		return;

	for (iter = netapplet->accesspoints; iter != NULL; iter = iter->next) {
		Accesspoint *ap = iter->data;

		if (strcmp (ap->essid, args[1]) == 0)
			netapplet->active_ap = ap;
	}

out:
	populate_popup_menu ();
}

static void
netapplet_handle_active (char **args)
{
	GSList *iter;
	gboolean changed;
	gboolean reconnected = FALSE;
	changed = FALSE;

	if (!netapplet->active ||
	    strcmp (netapplet->active->interface, args[1]) != 0) {

		netapplet->active = NULL;
		for (iter = netapplet->connections; iter != NULL;
		     iter = iter->next) {
			Connection *conn = iter->data;
			
			if (strlen(args[1]) && strcmp (conn->interface, args[1]) == 0) {
				netapplet->active = conn;
				changed = TRUE;
			}
		}
	}

	if (disconnected && netapplet->active) 
		reconnected = TRUE;

	if (!netapplet->active) {
		/* Nothing for this setup.  Fake it. */
		netapplet_set_icon (DISCONNECT_ICON);
		netapplet_free_accesspoints ();
		populate_popup_menu ();
		disconnected = TRUE;
		return;
	}

	
	if (g_str_has_prefix (netapplet->active->type, "Wireless")) {
		netapplet_get_accesspoints ();
		netapplet_get_wireless ();
		if (reconnected) {
			system("gnet setiface=nochange setconfig=nochange setap=nochange");
		}
	} else if (netapplet->accesspoints) {
		/* remove no-longer-valid wireless information */
		netapplet_free_accesspoints ();
	}

	if (changed) {
		netapplet_set_icon (netapplet->active->icon);
		populate_popup_menu ();
	}
}

static void
size_changed_cb (void)
{
	if (netapplet->icon_name == NULL)
		return;

	/* If the panel changed size, we might need to change the icon. */
	netapplet_set_icon (netapplet->icon_name);
}

static NetApplet *
netapplet_new (void)
{
	NetApplet *applet;

	/* Widgetry */
	applet = g_new0 (NetApplet, 1);
	applet->tray_icon = egg_tray_icon_new ("applet");

	applet->button = gtk_toggle_button_new ();
	gtk_button_set_relief (GTK_BUTTON (applet->button),
			       GTK_RELIEF_NONE);
	gtk_container_add (GTK_CONTAINER (applet->tray_icon),
			   applet->button);
	gtk_widget_show (applet->button);

	g_signal_connect (applet->button, "toggled",
			  G_CALLBACK (button_toggled_cb), NULL);
	g_signal_connect (applet->button, "size-allocate",
			  G_CALLBACK (size_changed_cb), NULL);

	applet->icon = gtk_image_new ();
	gtk_container_add (GTK_CONTAINER (applet->button), applet->icon);
	gtk_widget_show (applet->icon);

	applet->popup_menu = gtk_menu_new ();
	g_signal_connect (applet->popup_menu, "deactivate",
			  G_CALLBACK (menu_deactivate_cb), NULL);

	return applet;
}

int
main (int argc, char *argv[])
{
	GnomeClient *client;

	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	gnome_program_init ("netapplet", VERSION,
			    LIBGNOMEUI_MODULE,
			    argc, argv,
			    GNOME_PROGRAM_STANDARD_PROPERTIES,
			    GNOME_CLIENT_PARAM_SM_CONNECT, TRUE,
			    NULL);

	gtk_icon_theme_append_search_path (gtk_icon_theme_get_default (),
					   ICONDIR);

	client = gnome_master_client ();
	gnome_client_set_restart_command (client, argc, argv);
	if (netapplet_get_clipboard ())
		gnome_client_set_restart_style (client,
						GNOME_RESTART_IF_RUNNING);
	else {
		/* if we are already running, silently exit */
		gnome_client_set_restart_style (client, GNOME_RESTART_NEVER);
		return 1;
	}

	netapplet = netapplet_new ();
	netapplet_set_icon (DISCONNECT_ICON);
	populate_popup_menu ();

	netapplet_connect ();

	gtk_widget_show (GTK_WIDGET (netapplet->tray_icon));

	gtk_main ();

	return 0;
}
